/* 
 * wrapChan.c --
 *
 *	This file contains the code for the "rsrc" channel type. This kind
 *	of channel can read and uncompress on-the-fly from a zip-file.
 *      Assumed is that "rsrc" is an object from type "wrapRsrcType", which
 *	already contains the zip-file content-table in a hash-table.
 *
 *	The code for seting up the ZIP content table can be found in
 *	wrapRsrc.c
 *
 * Copyright (c) 1999      Jan Nijtmans
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <errno.h>
#include <string.h>
#include <time.h>
#ifdef __WIN32__
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif
#include "zlib.h"
#include "wrap.h"

#ifdef __WIN32__
#define Z_LIB_NAME "zlib.dll"
#endif

#ifndef Z_LIB_NAME
#define Z_LIB_NAME "libz.so"
#endif

#ifdef MAC_TCL
#  include "dlfcn.h"
#else
#  ifdef HAVE_DLFCN_H
#    include <dlfcn.h>
#  else
#    include "compat/dlfcn.h"
#  endif
#endif

/*
 * In some systems, like SunOS 4.1.3, the RTLD_NOW flag isn't defined
 * and this argument to dlopen must always be 1.
 */

#ifndef RTLD_NOW
#   define RTLD_NOW 1
#endif

#define BUF_SIZE (4096)

#define WRAP_SHORT(buf, num) (buf[num] + (buf[num+1]<<8))
#define WRAP_LONG(buf, num) (WRAP_SHORT(buf,num) + (buf[num+2]<<16) + (buf[num+3]<<24))

EXTERN Tcl_ObjType wrapRsrcType;

/*
 * Prototypes for procedures defined later in this file:
 */

static int	rsrcClose _ANSI_ARGS_((ClientData instanceData,
		   Tcl_Interp *interp));
static int	rsrcInput _ANSI_ARGS_((ClientData instanceData,
		    char *buf, int toRead, int *errorCodePtr));
static int	rsrcOutput _ANSI_ARGS_((ClientData instanceData,
	            char *buf, int toWrite, int *errorCodePtr));
static void	rsrcWatchChannel _ANSI_ARGS_((ClientData instanceData,
		    int mask));
static int	rsrcGetOption _ANSI_ARGS_((
		    ClientData instanceData, Tcl_Interp* interp,
		    char *optionName, Tcl_DString *dsPtr));
static void	rsrcChannelReady _ANSI_ARGS_((ClientData instanceData));
static int      rsrcGetFile      _ANSI_ARGS_((ClientData instanceData,
		    int direction, ClientData* handlePtr));
static int	loadLib _ANSI_ARGS_((Tcl_Interp *interp, CONST char *libName,
		    VOID **handlePtr, char **symbols, int num));

/*
 * This structure describes the channel type structure for rsrc channels:
 */

static Tcl_ChannelType channelType = {
  "rsrc",		/* Type name.                                    */
  NULL,			/* Set blocking/nonblocking behaviour. NULL'able */
  rsrcClose,		/* Close channel, clean instance data            */
  rsrcInput,		/* Handle read request                           */
  rsrcOutput,		/* Handle write request                          */
  NULL,			/* Move location of access point.      NULL'able */
  NULL,			/* Set options.                        NULL'able */
  rsrcGetOption,	/* Get options.                        NULL'able */
  rsrcWatchChannel,	/* Initialize notifier                           */
  rsrcGetFile		/* Get OS handle from the channel.               */
};

/*
 * This structure describes the per-instance state of a in-memory channel.
 */

typedef struct ChannelInstance {
  int crc;                  /* crc */
  time_t date;        	    /* time-date stamp */
  unsigned int length;	    /* number of bytes left. */
  int compression;          /* 0 = no compression */
			    /* 8 = deflate compression */
  Tcl_Channel    chan;      /* backreference to underlying channel */
  Tcl_Channel    rsrcChan;  /* backreference to generic channel information */

  unsigned int clength;     /* number of bytes left for compressor. */
  char *buffer;		    /* buffer for compressor */
  z_stream stream;	    /* stream for compressor */
} ChannelInstance;



/*
 * Sigh, ZLIB_VERSION is defined as a string so there's no
 * way to do a proper check here.  Instead we guess based
 * on the presence of defines that were added between the
 * 0.95 and 1.0 distributions.
 */
#if !defined(Z_NO_COMPRESSION) || !defined(Z_DEFLATED)
#error "Antiquated ZLIB software; you must use version 1.0 or later"
#endif

/*
 * Stuff to support dynamical loading of libz.
 * for this simple application, we need only 3 functions.
 */

static struct LibzFunctions {
    VOID *handle;
    int (* inflate) _ANSI_ARGS_((z_streamp, int));
    int (* iInit2) _ANSI_ARGS_((z_streamp, int, CONST char *, int));
    int (* inflateEnd) _ANSI_ARGS_((z_streamp));
} zlib = {0};

static char *symbols[] = {
    "inflate",
    "inflateInit2_",
    "inflateEnd",
    (char *) NULL
};


/*
 *-------------------------------------------------------------------------
 *
 * Wrap_Open --
 *
 *	This procedure opens a channel to the resource.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 * *-------------------------------------------------------------------------
 */

Tcl_Channel Wrap_Open(interp, rsrc, name)
    Tcl_Interp *interp;
    Tcl_Obj *rsrc;
    CONST char *name;
{
    static unsigned long rsrcCounter = 0;
    ChannelInstance *priv;
    Tcl_Channel chan, rsrcChan;
    unsigned char buffer[1024];
    char *string;
    int pos, i;
    Tcl_HashEntry *entry;
    struct tm tm;

    if (Tcl_ConvertToType(interp, rsrc, &wrapRsrcType) != TCL_OK) {
	return NULL;
    }

    /*
     *    Now we can be sure that rsrc is fully initialized.
     *    This object uses the string representation to store
     *    the file name, and the otherValuePtr to store a
     *    hash table of offset position.
     *    Opening this resource is simple now:
     *    - Open a channel to the filename in binary mode
     *    - Search for the entry name in the hashtable.
     *      The hash value is the required offset in the file.
     *    - Seek to the found position. Here a table can be
     *      found with all other required information.
     */

    string = Tcl_GetString(rsrc);
    chan = Tcl_OpenFileChannel(interp, string, "r", 0);
    if (!chan) {
	return NULL;
    }
    if (Tcl_SetChannelOption(interp, chan, "-translation", "binary") != TCL_OK) {
	/* this should never happen */
	Tcl_Close(NULL, chan);
	return NULL;
    }
    if (Tcl_SetChannelOption(interp, chan, "-encoding", "binary") != TCL_OK) {
	/* this should never happen */
	Tcl_Close(NULL, chan);
	return NULL;
    }
    entry = Tcl_FindHashEntry((Tcl_HashTable *) rsrc->internalRep.otherValuePtr, name);
    if (!entry) {
    	Tcl_Close(interp, chan);
	Tcl_AppendStringsToObj(Tcl_GetObjResult(interp), "entry \"",
		name, "\" not found in resource \"",
		string,"\"", NULL);
 	return NULL;
    }
    pos = (int) Tcl_GetHashValue(entry);

    Tcl_Seek(chan, pos, SEEK_SET);
    Tcl_Read(chan, buffer, 30);
    if (memcmp(buffer, "\120\113\03\04", 4)) {
	if (interp) {
	    Tcl_AppendResult(interp, "local header mismatch", NULL);
	}
    	Tcl_Close(interp, chan);
	return NULL;
    }

    priv =  (ChannelInstance*) ckalloc (sizeof (ChannelInstance));

    if ((priv->compression = (buffer[6]&2) ? WRAP_SHORT(buffer,8):0)) {
	z_stream *stream = &priv->stream;
	if (loadLib(interp, Z_LIB_NAME, &zlib.handle, symbols, 3) != TCL_OK) {
	    Tcl_Close(interp, chan);
	    return NULL;	    
	}
	priv->buffer = (unsigned char *) ckalloc(BUF_SIZE);
	stream->zalloc = (alloc_func)0;
	stream->zfree = (free_func)0;
	stream->opaque = (voidpf)0;
	stream->avail_in = 0;

	zlib.iInit2(&(priv->stream), -MAX_WBITS, ZLIB_VERSION, sizeof(z_stream));

    } else {
	priv->buffer = 0;
    }

    i = WRAP_SHORT(buffer,10);
    tm.tm_sec = (i & 0x1F) * 2;
    tm.tm_min = (i >> 5) & 0x3F;
    tm.tm_hour = (i >> 11) & 0x1F;
    i = WRAP_SHORT(buffer,12);
    tm.tm_mday = i & 0x1F;
    tm.tm_mon = ((i >> 5) & 0xF) - 1;
    tm.tm_year = ((i >> 9) & 0xFF) + 80;
    tm.tm_wday = 0;
    tm.tm_yday = 0;
    tm.tm_isdst = -1;

    priv->date = mktime(&tm);
    priv->crc = WRAP_LONG(buffer,14);
    priv->clength = WRAP_LONG(buffer,18);
    priv->length = WRAP_LONG(buffer,22);
    Tcl_Seek(chan, WRAP_SHORT(buffer,26)+WRAP_SHORT(buffer,28), SEEK_CUR);

    sprintf((char *) buffer, "rsrc%lu", rsrcCounter++);
    rsrcChan = Tcl_CreateChannel (&channelType,
	    (char *) buffer, (ClientData) priv, TCL_READABLE);
    priv->rsrcChan  = rsrcChan;
    priv->chan      = chan;
    Tcl_AppendResult(interp, (char *) buffer, (char*) NULL);
    return rsrcChan;
}


static int
rsrcClose (instanceData, interp)
    ClientData  instanceData;    /* The instance information of
    				  * the channel to close */
    Tcl_Interp* interp;
{
    int result;
    ChannelInstance* priv = (ChannelInstance*) instanceData;

    if (priv->buffer != (char*) NULL) {
	ckfree ((char*) priv->buffer);
	zlib.inflateEnd(&(priv->stream));
    }
    result = Tcl_Close(interp, priv->chan);
    ckfree ((char*) priv);

    return result;
}



static int
rsrcInput (instanceData, buf, toRead, errorCodePtr)
    ClientData instanceData;	/* The channel to read from */
    char*      buf;			/* Buffer to fill */
    int        toRead;		/* Requested number of bytes */
    int*       errorCodePtr;	/* Location of error flag */
{
    ChannelInstance* priv = (ChannelInstance*) instanceData;

    if (toRead > priv->length) {
	toRead = priv->length;
    }

    if (toRead == 0)
	return 0;

    if (priv->compression) {
	int err = Z_OK;
	z_stream *stream = &priv->stream;
	stream->next_out = buf;
	stream->avail_out = toRead;
	while (stream->avail_out) {
	    if (!stream->avail_in) {
		int len = priv->clength;
		if (len > BUF_SIZE) {
		    len = BUF_SIZE;
		}
		len = Tcl_Read(priv->chan, priv->buffer, len);
		priv->clength -= len;
		stream->next_in = priv->buffer;
		stream->avail_in = len;
	    }
	    err = zlib.inflate(stream, Z_NO_FLUSH);
	    if (err) break;
	}
	if (err == Z_STREAM_END) {
	    if ((stream->avail_out != 0)) {
		*errorCodePtr = err; /* premature end */
		return -1;
	    }
	} else if (err) {
	    *errorCodePtr = err; /* some other zlib error */
	    return -1;
	}
    } else {
	toRead = Tcl_Read(priv->chan, buf, toRead);
    }
    priv->length -= toRead;

    *errorCodePtr = 0;
    return toRead;
}


static int
rsrcOutput (instanceData, buf, toWrite, errorCodePtr)
    ClientData instanceData;	/* The channel to write to */
    char*      buf;		/* Data to be stored. */
    int        toWrite;		/* Number of bytes to write. */
    int*       errorCodePtr;	/* Location of error flag. */
{
    *errorCodePtr = EINVAL;
    return -1;
}

static void
rsrcWatchChannel (instanceData, mask)
    ClientData instanceData;	/* Channel to watch */
    int        mask;		/* Events of interest */
{
    /* not implemented yet. Do we need it???? */
}

static int
rsrcGetOption (instanceData, interp, optionName, dsPtr)
    ClientData   instanceData;	/* Channel to query */
    Tcl_Interp*  interp;	/* Interpreter to leave error messages in */
    char*        optionName;	/* Name of reuqested option */
    Tcl_DString* dsPtr;		/* String to place the result into */
{
    ChannelInstance* priv;
    long result;
    char buffer[50];
    buffer[0] = 0;

    priv = (ChannelInstance*) instanceData;

    if (!optionName || !strcmp(optionName, "-crc")) {
	if (!optionName) {
	    Tcl_DStringAppendElement (dsPtr, "-crc");
	}
	sprintf (buffer, "%d", priv->crc);
	Tcl_DStringAppendElement (dsPtr, buffer);
    }
    if (!optionName || !strcmp(optionName, "-date")) {
	if (!optionName) {
	    Tcl_DStringAppendElement (dsPtr, "-date");
	}
	sprintf (buffer, "%ul", (unsigned long) priv->date);
	Tcl_DStringAppendElement (dsPtr, buffer);
    }
    if (!optionName || !strcmp(optionName, "-length")) {
	if (!optionName) {
	    Tcl_DStringAppendElement (dsPtr, "-length");
	}
	sprintf (buffer, "%u", priv->length);
	Tcl_DStringAppendElement (dsPtr, buffer);
    }

    if (!buffer[0]) {
	Tcl_SetErrno (EINVAL);
	return Tcl_BadChannelOption (interp, optionName, "crc date length");
    }
    return TCL_OK;
}

static int
rsrcGetFile (instanceData, direction, handlePtr)
  ClientData  instanceData;	/* Channel to query */
  int         direction;	/* Direction of interest */
  ClientData* handlePtr;	/* Space to the handle into */
{
    return TCL_ERROR;
}



/*
 *----------------------------------------------------------------------
 *
 * loadLib --
 *
 *	This procedure is called to load a shared library into memory.
 *	If the extension is ".so" (e.g. Solaris, Linux) or ".sl" (HP-UX)
 *	it is possible that the extension is appended or replaced with
 *	a major version number. If the file cannot be found, the version
 *	numbers will be stripped off one by one. e.g.
 *
 *	HP-UX:	libtiff.3.4	Linux,Solaris:	libtiff.so.3.4
 *		libtiff.3			libtiff.so.3
 *		libtiff.sl			libtiff.so
 *
 * Results:
 *	TCL_OK if function succeeds. Otherwise TCL_ERROR while the
 *	interpreter will contain an error-message. The last parameter
 *	"num" contains the minimum number of symbols that is required
 *	by the application to succeed. Only the first <num> symbols
 *	will produce an error if they cannot be found.
 *
 * Side effects:
 *	At least <num> Library functions become available by the
 *	application.
 *
 *----------------------------------------------------------------------
 */

typedef struct Functions {
    VOID *handle;
    int (* first) _ANSI_ARGS_((void));
    int (* next) _ANSI_ARGS_((void));
} Functions;

#define LOAD_FAILED ((VOID *) -114)
/*
 * Macro to use to fill in "offset" fields of Tk_ConfigInfos.
 * Computes number of bytes from beginning of structure to a
 * given field.
 */

#ifndef offsetof
#define offsetof(type, field) ((char *) &((type *) 0)->field)
#endif

static int
loadLib(interp, libName, handlePtr, symbols, num)
    Tcl_Interp *interp;
    CONST char *libName;
    VOID **handlePtr;
    char **symbols;
    int num;
{
    VOID *handle = (VOID *) NULL;
    Functions *lib = (Functions *) handlePtr;
    char **p = (char **) &(lib->first);
    char **q = symbols;
    char buf[256];
    char *r;
    int length;

    if (lib->handle != NULL) {
	return (lib->handle != LOAD_FAILED) ? TCL_OK : TCL_ERROR;
    }
    length = strlen(libName);
    strcpy(buf,libName);
    handle = dlopen(buf, RTLD_NOW);

    while (handle == NULL) {
	if ((r = strrchr(buf,'.')) != NULL) {
	    if ((r[1] < '0') || (r[1] > '9')) {
		if (interp) {
		    Tcl_AppendResult(interp,"cannot open ",libName,
			    ": ", dlerror(), (char *) NULL);
		} else {
		    printf("cannot open %s: %s\n",libName,dlerror());
		}
		lib->handle = LOAD_FAILED;
		return TCL_ERROR;
	    }
	    length = r - buf;
	    *r = 0;
	}
	if (strchr(buf,'.') == NULL) {
	    strcpy(buf+length,".sl");
	    length += 3;
	}
	dlerror();
	handle = dlopen(buf, RTLD_NOW);
    }
    buf[0] = '_';
    while (*q) {
	*p = (char *) dlsym(handle,*q);
	if (*p == (char *)NULL) {
	    strcpy(buf+1,*q);
	    *p = (char *) dlsym(handle,buf);
	    if ((num > 0) && (*p == (char *)NULL)) {
		if (interp) {
		    Tcl_AppendResult(interp,"cannot open ",libName,
			    ": symbol \"",*q,"\" not found", (char *) NULL);
		} else {
		    printf("cannot open %s: symbol \"%s\" not found",
			    libName, *q);
		}
		dlclose(handle);
		lib->handle = LOAD_FAILED;
		return TCL_ERROR;
	    }
	}
	q++; num--;
	p += (((int) offsetof(Functions, next)) - ((int) offsetof(Functions, first))) /
		sizeof(char *);
    }
    lib->handle = handle;

    return TCL_OK;
}
